#!/bin/bash

# LibreGates/wrap.sh
# created sam. sept. 12 09:29:47 CEST 2020
#
# Generates a wrapper from the XML output of GHDL

ReadTag () {
  IFS=\> read -d \< ENTITY
}

GetAttribute () {      #input list in $1
  LIST=($ATTRIBUTES)   # split string into array
  ATTRIBUTE=""
  for i in $(seq "${#LIST[@]}") ; do  # scan the list
    IFS='"' read -a ATTRIBUTE <<< "${LIST[$i]}"  # split the individual parameter in a sub-string
    #echo "  $ATTRIBUTE   -   ${ATTRIBUTE[1]}"
    if [[ $ATTRIBUTE = "$1" ]] ; then   # found the key
      echo -n "${ATTRIBUTE[1]}"         # return the value
    fi
  done
}

PORT_TYPE=""
PORTNAME=""
INOUT=""

FlushPort () {
  if [[ "$PORT_TYPE" != "" ]] ; then
    case "$INOUT" in
      "in")  ;;
      "out") ;;
      *) echo "wrong port direction for signal $PORTNAME"
         return
    esac
    echo "$PORTNAME $PORT_TYPE $INOUT"
  fi
}

# keep only the first "port_chain" block which should be the top-level interface
ghdl --file-to-xml $*        |  # transform the VHDL into XML
sed  '/ <\/port_chain>$/,$d' |  # drop everything after the first end of block
sed -n '/ <port_chain>$/,$p' |  # drop everything before the first start of block
while ReadTag; do               # parse the stream
  TAG_NAME=${ENTITY%% *}
  ATTRIBUTES=${ENTITY#* }

  # executed in sequence :
  if [[ $TAG_NAME = "el" ]] ; then
    FlushPort
    LIST=($ATTRIBUTES)               # split string into array
    #    echo "- $TAG_NAME : ${#LIST[@]}" # display array size
    INOUT=$(GetAttribute 'mode=')
    PORTNAME=$(GetAttribute 'identifier=')
  fi
  if [[ $TAG_NAME = "subtype_indication" ]] ; then
    PORT_TYPE=$(GetAttribute 'identifier=')
  fi
done

FlushPort

